RESOLVE_PREFIX = 'resolve_'
POST_PREFIX = 'post_'
PYDANTIC_FORWARD_REF_UPDATED = '__pydantic_resolve_forward_refs_updated__'
DATACLASS_FORWARD_REF_UPDATED = '__pydantic_resolve_dataclass_forward_refs_updated__'
HAS_MAPPER_FUNCTION = '__pydantic_resolve_mapper_provided__'
POST_DEFAULT_HANDLER = 'post_default_handler'
EXPOSE_TO_DESCENDANT = '__pydantic_resolve_expose__'
COLLECTOR_CONFIGURATION = '__pydantic_resolve_collect__'
