class ResolverTargetAttrNotFound(Exception):
    pass

class LoaderFieldNotProvidedError(Exception):
    pass

class MissingAnnotationError(Exception):
    pass

class GlobalLoaderFieldOverlappedError(Exception):
    pass

class MissingCollector(Exception):
    pass